<?php declare(strict_types=1);

namespace Lewisdale\Webmentions;

use League\Uri\Exceptions\SyntaxError;
use Lewisdale\Webmentions\Exceptions\InvalidTargetException;
use Lewisdale\Webmentions\Exceptions\InvalidUrlException;
use Lewisdale\Webmentions\Gateways\WebmentionGatewayInterface;
use Lewisdale\Webmentions\Models\Webmention;
use Symfony\Component\HttpClient\HttpClient;
use League\Uri\Uri;
use Lewisdale\Webmentions\Exceptions\SourceNotFoundException;
use Lewisdale\Webmentions\Exceptions\TargetNotMentionedException;
use Lewisdale\Webmentions\Models\MentionType;
use Symfony\Component\DomCrawler\Crawler;
use Symfony\Contracts\HttpClient\HttpClientInterface;

class Endpoint {

    function __construct(
        private readonly HttpClientInterface $httpClient,
        private readonly WebmentionGatewayInterface $gateway
    )
    {}

    public function validateUrl(string $url) : bool {
        try {
            $uri = Uri::createFromString($url);
            $scheme = $uri->getScheme();
            $schemeValid = in_array($scheme, ["http", "https"]);

            return $schemeValid && !!filter_var($url, FILTER_VALIDATE_URL);
        
        } catch (SyntaxError $e)
        {
            return false;
        }
    }
    public function receiveWebmention(string $source, string $target) : void
    {
        // Validate that both source and target are actual domains
        if (!$this->validateUrl($source) || !$this->validateUrl($target))
        {
            throw new InvalidUrlException();
        }

        // Validate that the webmention target is a domain I care about
        if (!str_contains($target, "https://lewisdale.dev")) {
            throw new InvalidTargetException();
        }

        // Parse content from the source
        $response = $this->httpClient->request('GET', $source);

        if ($response->getStatusCode() < 400)
        {
            $document = new Crawler($response->getContent());

            if (!$this->hasMention($target, $document))
            {
                throw new TargetNotMentionedException();
            }

            $container = $this->getContainer($target, $document);
            $type = $this->parseMentionType($target, $container);
            $content = $this->parseContent($target, $container, $type);
            $author = $this->parseAuthor($container);

            $webmention = new Webmention(null, $target, $source, $type, null, $author);
            $this->gateway->save($webmention);
        } else {
            throw new SourceNotFoundException();
        }
    }

    private function hasMention(string $target, Crawler $document) : bool
    {
        return $document->filter('a[href="' . $target . '"]')->count() > 0;
    }

    private function getContainer(string $target, Crawler $document) : Crawler
    {
        return $document->filter('a[href="' . $target . '"]')->closest('.h-entry') ?? $document;
    }

    private function parseMentionType(string $target, Crawler $document) : MentionType
    {
        $class = $document->filter('a[href="'. $target . '"]')->attr('class');

        if (str_contains($class, "u-like-of")) {
            return MentionType::Like;
        } else if (str_contains($class, "u-in-reply-to")) {
            return MentionType::Reply;
        } else if (str_contains($class, "u-repost-of")) {
            return MentionType::Repost;
        }

        return MentionType::Mention;
    }

    private function parseContent(string $target, Crawler $document, MentionType $type) : ?string
    {
        return match ($type) {
            MentionType::Like => "Liked this post",
            MentionType::Reply => $document->innerText(),
            MentionType::Repost => "Reposted this post",
            MentionType::Mention => $document->innerText(),
        };
    }

    private function parseAuthor(Crawler $document) : ?string
    {
        $card = $document->filter('.p-author.h-card')->eq(0);
        
        if ($card->count())
        {
            $name = $card->filter('.p-name')?->text("");
            $url = $card->filter('.u-url')?->text("");
            $photo = $card->filter('.u-photo')->count() ? $card->filter('.u-photo')->attr('src') : "";

            return implode(", ", [$name, $url, $photo]);
        }
        return null;
    }
}